clear;

%Read the calibrated gyroscope, accelerometer and magnetometer measurements
u=csvread('Data/XsensAccGyrMag2.csv');

%Timestep of the ground truth and measurements
Ts=1/100;

%Set fontsize for plots
fontsize=14;

%Run the zero velocity detector
gamma=3e5;
sigma_a=1;
sigma_g=0.7e-06;
zupt=zero_velocity_detector(u,gamma,sigma_a,sigma_g);

%Limit u, gt, and zupt detect
N=500;
u=u(:,1:N);
zupt=zupt(:,1:N);
N=size(u,2);

%Calculate ZUPT-EKF position estimates
[p_odom, v_odom, eta_odom, P_odom]=ZUPTEKF(u,zupt,Ts);

%Generate odometry based on ZUPT-ekf position estimates
delta_p=p_odom(:,2:N)-p_odom(:,1:(N-1));

%Number of particles
M=50;

%Pre-allocate position trajectory for all particles
p=zeros(3,N,M);

%Pre-allocate position trajectory for loosely coupled dynamic update
p_loose=zeros(3,N,M);

%Initialize velocity and orientation states and covariances
[Q,R_v,v,P,P_diag,R,quat,acc_t]=ZUPTEKFInit(u,Ts,N,M);

%Set covariance of position dynamics white noise
Q_p=0.0000*eye(3);

%Pre-allocate model deviation estimates
e_tight=zeros(N,M);
e_loose=zeros(N,M);

%iterate through all timesteps
for t=2:N
    
    %PF Dynamic update tight
    w_MC=zeros(3,M);
    for i=1:M
        w_MC(:,i)=mvnrnd(zeros(3,1),P(1:3,1:3,i))'+1./Ts*mvnrnd(zeros(3,1),Q_p)';
        p(:,t,i)=p(:,t-1,i)+Ts*(v(:,t-1,i)+w_MC(:,i));
    end
    
    %KF Dynamic measurement update
    [v,P,P_diag,R,quat,epsilon]=KFDYNMEASUPDATE(v,P,P_diag,R,quat,Q_p,Ts,t-1,w_MC,M);
    
    %KF Dynamic update
    [v,P,P_diag,R,quat,acc_t]=ZUPTEKFDynUpdate(v,P,P_diag,R,quat,acc_t,u,Ts,t,M,Q);
    
    
    if zupt(t)
        [v,P,P_diag,R,quat]=ZUPTEKFMeasUpdate(v,P,P_diag,R,quat,t,M,R_v);
    end
    
    %PF Dynamic update loose
    if (mod(t,10)==0)
        if zupt(t)==0
            %perform dynamic update of all particles
            for i=1:M
                %Dynamic update
                noise=mvnrnd(zeros(3,1),diag([0.0005 0.0005 0.001]))';
                p_loose(:,t,i)=p_loose(:,t-1,i)+delta_p(:,t-1)+noise;
            end
        else
            for i=1:M
                p_loose(:,t,i)=p_loose(:,t-1,i)+delta_p(:,t-1);
            end
        end
    else
        for i=1:M
            p_loose(:,t,i)=p_loose(:,t-1,i)+delta_p(:,t-1);
        end
    end
    
end

figure(1); clf;
subplot(1,2,1);
plot3(p(1,:,1),p(2,:,1),p(3,:,1)+5,'k');
hold on;
for i=2:M
    plot3(p(1,:,i),p(2,:,i),p(3,:,i),'Color',[0.9,0.9,0.9]);
end
xlabel('$x$/[m]','Interpreter','Latex','Fontsize',fontsize);
ylabel('$y$/[m]','Interpreter','Latex','Fontsize',fontsize);
zlabel('$z$/[m]','Interpreter','Latex','Fontsize',fontsize);
title('Tightly coupled Mag-SLAM','Interpreter','Latex','Fontsize',fontsize);
axis equal;
xlim([-2.1 0.1]);
ylim([-3.3 0.2]);
view(2);
grid on;
subplot(1,2,2);
plot3(p_loose(1,:,1),p_loose(2,:,1),p_loose(3,:,1)+5,'k');
hold on;
for i=2:M
    plot3(p_loose(1,:,i),p_loose(2,:,i),p_loose(3,:,i),'Color',[0.9,0.9,0.9]);
    hold on;
end
xlabel('$x$/[m]','Interpreter','Latex','Fontsize',fontsize);
ylabel('$y$/[m]','Interpreter','Latex','Fontsize',fontsize);
zlabel('$z$/[m]','Interpreter','Latex','Fontsize',fontsize);
title('Loosely coupled Mag-SLAM','Interpreter','Latex','Fontsize',fontsize);
axis equal;
xlim([-2.1 0.1]);
ylim([-3.3 0.2]);
grid on;
view(2);

