clear;

%Read the ground truth position estimate
gt=csvread('Data/fake_ground_truth_Xsens2.csv');

%Read the calibrated gyroscope, accelerometer and magnetometer measurements
u=csvread('Data/XsensAccGyrMag2.csv');

%Timestep of the ground truth and measurements
Ts=1/100;

%Set plot font-size
fontsize=14;

%Set room dimensions
margin=2; %distance from trajectory to wall [m]
uL1=max(gt(1,:))+margin; %upper limit x, a.k.a. the position of the right wall [m]
lL1=min(gt(1,:))-margin; %lower limit x, a.k.a. the position of the left wall [m]
uL2=max(gt(2,:))+margin; %uppper limit y, a.k.a. the position of the upper wall [m]
lL2=min(gt(2,:))-margin; %lower limit y, a.k.a. the position of the lower wall [m]
uL3=max(gt(3,:))+margin; %upper limit z, a.k.a. the position of the ceiling [m]
lL3=min(gt(3,:))-margin; %lower limit z, a.k.a. the position of the floor [m]

%Run the zero velocity detector
gamma=3e5;
sigma_a=1;
sigma_g=0.7e-06;
zupt=zero_velocity_detector(u,gamma,sigma_a,sigma_g);

%Calculate ZUPT-EKF position estimates
p_odometry=ZUPTEKF(u,zupt,Ts);

%Downsample and limit u, gt, zupt detect and odometry
N=10000;
u=u(:,1:10:N);
p_odometry=p_odometry(:,1:10:N);
zupt=zupt(:,1:10:N);
gt=gt(:,1:10:N); Ts=1/10; N=size(gt,2);

%Set up movie object
vidObj = VideoWriter('Particle_filter_3D.avi');
vidObj.Quality = 100;
vidObj.FrameRate = Ts*N;
open(vidObj);
set(gca,'nextplot','replacechildren');
set(gca, 'visible', 'off');
F(N) = struct('cdata',[],'colormap',[]);

%Generate odometry based on ZUPT-ekf position estimates
delta_p=p_odometry(:,2:N)-p_odometry(:,1:(N-1));

%Define hyperparameters for the magnetic field map GP prior
sigma_SE=1;
l_SE=0.3;
sigma_y=0.1;

%Calculate the magnetic field norm measurements from the
%magnetic field measurements
y_m=zeros(size(gt,2),1);
for i=1:N
    y_m(i)=norm(u(7:9,i));
end

%Center the magnetic field norm measurements around zero
mean_y_m=mean(y_m);
y_m=y_m-mean_y_m;

%Number of basis functions used in Reduced-Rank approximation
N_m=2000;

%Calculate Lambda and the order of indices used in the
%analytic basis functions of the Reduced-Rank Approximation
[Indices, Lambda]=Lambda3D(N_m,uL1,lL1,uL2,lL2,uL3,lL3,sigma_SE,l_SE);

%Number of particles
M=100;

%Pre-allocate position trajectory for all particles
p=zeros(3,N,M);

%Initialise all weights to 1/M
weights=ones(1,M)./M;

%Pre-allocate magnetic field state vector for all particles
m=zeros(N_m,M);

%Pre-allocate the magnetic field variance for all particles
P_m=zeros(N_m,N_m,M);
for i=1:M
    P_m(:,:,i)=Lambda;
end

%iterate through all timesteps
for t=2:N
    
    if zupt(t)==0
        
        %perform dynamic update of all particles
        for i=1:M
            %Dynamic update
            p(:,t,i)=p(:,t-1,i)+delta_p(:,t-1)+mvnrnd(zeros(3,1),diag([0.0005 0.0005 0.001]))';
        end
        
        %Perform measurement update of all particles
        delay=20;
        
        if t>delay
            
            for i=1:M
                
                %Update the magnetic field maps based on the incoming
                %measurements - Kalman filter measurement update
                phi_t=Phi3D(p(:,t-delay,i),N_m,uL1,lL1,uL2,lL2,uL3,lL3,Indices);
                S_t=phi_t*P_m(:,:,i)*phi_t'+sigma_y^2;
                K_t=P_m(:,:,i)*phi_t'./S_t;
                m(:,i)=m(:,i)+K_t*(y_m(t-delay)-phi_t*m(:,i));
                P_m(:,:,i)=P_m(:,:,i)-K_t*S_t*K_t';
                P_m(:,:,i)=1/2*(P_m(:,:,i)+P_m(:,:,i)');
                
                
                
                %Evaluate weights of particles
                phi_s=Phi3D(p(:,t,i),N_m,uL1,lL1,uL2,lL2,uL3,lL3,Indices);
                mu=phi_s*m(:,i);
                var=phi_s*P_m(:,:,i)*phi_s'+sigma_y^2;
                weights(i) = normpdf(y_m(t),mu,sqrt(var))*weights(i);
                
            end
            
            %Normalise particle weights
            weights=weights./sum(weights);
            
            %Selective resampling of particles
            M_eff=1/sum(weights.^2);
            if (M_eff<(2*M/3))
                [p,m,P_m]=resample(p,m,P_m,weights);
                weights=0*weights+1./M;
            end
        end
        
        %Make a new invisible figure for every frame
        h=figure('Visible','off'); clf;
        
        %Find the particle number with the highest weight
        [max_val, arg_max]=max(weights);
        
        %Plot the magnetic field estimate,
        %defined as the magnetic field of the particle
        %that currently has the highest weight
        plot_projection_mag_field_map3D(m(:,arg_max),P_m(:,:,arg_max),Indices,uL1,lL1,uL2,lL2,uL3,lL3,0,0.1,mean_y_m,fontsize)
        hold on;
        
        %Plot the position trajectory estimate,
        %defined as the trajectory of the particle
        %that currently has the highest weight
        plot(p(1,1:t,arg_max),p(2,1:t,arg_max),'k','linewidth',1.5);
        xlim([lL1 uL1]);
        ylim([lL2 uL2]);
        scatter(p(1,t,:),p(2,t,:),1,'ko');
        
        %Save selected frames to full-resolution figures
        if (t==24 || t==192 || t==380 || t==997)
            savefig(h,['Cache/picture_at_timestep_',num2str(t),'.fig']);
        end
        
        %Store plot to video
        F(t) = getframe;
        currFrame = getframe(gcf);
        writeVideo(vidObj,currFrame);
    else
        for i=1:M
            p(:,t,i)=p(:,t-1,i)+delta_p(:,t-1);
        end
    end
    
end

close(vidObj);

RMSE_PF=calculate_RMSE(p,weights,5,zupt);
disp(['RMSE RBPF: ',num2str(RMSE_PF)]);
RMSE_ODOMETRY=calculate_RMSE(p_odometry,1,6,zupt);
disp(['RMSE ZUPT EKF: ',num2str(RMSE_ODOMETRY)]);

figure(1); clf;
plot3(p(2,:,arg_max),p(1,:,arg_max),p(3,:,arg_max),'k');
hold on;
plot3(p_odometry(2,:),p_odometry(1,:),p_odometry(3,:),'b');
xlabel('$x/[m]$','Interpreter','Latex','Fontsize',fontsize);
ylabel('$y/[m]$','Interpreter','Latex','Fontsize',fontsize);
zlabel('$z/[m]$','Interpreter','Latex','Fontsize',fontsize);
legend({'Loosely coupled Mag-SLAM','Zero-velocity aided EKF'},'Interpreter','Latex','Fontsize',fontsize);
grid on;

[max_val arg_max]=max(weights);
figure(2); clf;
plot(Ts:Ts:(N*Ts),p(1,:,arg_max),'k');
hold on;
plot(Ts:Ts:(N*Ts),p_odometry(1,:),'b');
xlabel('t/[s]','Interpreter','Latex','Fontsize',fontsize);
ylabel('x/[m]','Interpreter','Latex','Fontsize',fontsize);
legend({'Loosely coupled Mag-SLAM','Zero-velocity aided EKF'},'Interpreter','Latex','Fontsize',fontsize);

figure(3); clf;
plot(Ts:Ts:(N*Ts),p(2,:,arg_max),'k');
hold on;
plot(Ts:Ts:(N*Ts),p_odometry(2,:),'b');
xlabel('t/[s]','Interpreter','Latex','Fontsize',fontsize);
ylabel('y/[m]','Interpreter','Latex','Fontsize',fontsize);
legend({'Loosely coupled Mag-SLAM','Zero-velocity aided EKF'},'Interpreter','Latex','Fontsize',fontsize);

figure(4); clf;
plot(Ts:Ts:(N*Ts),p(3,:,arg_max),'k');
hold on;
plot(Ts:Ts:(N*Ts),p_odometry(3,:),'b');
xlabel('t/[s]','Interpreter','Latex','Fontsize',fontsize);
ylabel('z/[m]','Interpreter','Latex','Fontsize',fontsize);
legend({'Loosely coupled Mag-SLAM','Zero-velocity aided EKF'},'Interpreter','Latex','Fontsize',fontsize);

openfig('Cache/picture_at_timestep_24.fig','new','visible');
openfig('Cache/picture_at_timestep_192.fig','new','visible');
openfig('Cache/picture_at_timestep_380.fig','new','visible');
openfig('Cache/picture_at_timestep_997.fig','new','visible');

