%Use repeated simulations to investigate the convergence of the PF using
%various biases and process noise

%Settings for simulations
experiments=50;
noises=(0:7).^2;
n_noises=length(noises);

%Pre-allocate simulation results
RMSEsODOMETRY=zeros(experiments,n_noises);
RMSEsRBPF=zeros(experiments,n_noises);

%Set room dimensions
margin=2; %distance from trajectory to wall [m]
uL1=4; %upper limit x, a.k.a. the position of the right wall [m]
lL1=-4; %lower limit x, a.k.a. the position of the left wall [m]
uL2=5; %uppper limit y, a.k.a. the position of the upper wall [m]
lL2=-5; %lower limit y, a.k.a. the position of the lower wall [m]
uL3=2; %upper limit z, a.k.a. the position of the ceiling [m]
lL3=-2; %lower limit z, a.k.a. the position of the floor [m]

%Simulate the ground truth position estimate
Ts=1/10; %Sampling time [s]
v=1.4; %Walking speed [m/s]
z=0; %z position of trajectory
gt=rectangular_walk_ground_truth(Ts,lL1,uL1,lL2,uL2,margin,v,z);
N=size(gt,2);

%Number of basis functions used in Reduced-Rank approximation
N_m=500;

%Variance of odometry measurement noise
sigma_p=0.01;

%Simulated odometry measurement bias
b=[0; 0.005; 0];

%Number of particles
M=100;

%Define hyperparameters for the magnetic field map GP prior
sigma_SE=1;
l_SE=0.6;
sigma_y=0.1;

for experiment=1:experiments
    
    %Simulate magnetic field norm in the trajectory
    mag_field_norm=mvnrnd(zeros(size(gt,2),1),Kern(gt,gt,sigma_SE,l_SE));
    
    %Simulate magnetic field norm measurements in the trajectory
    y_sim=mag_field_norm+mvnrnd(0,sigma_y^2,size(gt,2))';
    
    %Calculate Lambda and the order of indices used in the
    %analytic basis functions of the Reduced-Rank Approximation
    [Indices, Lambda]=Lambda3D(N_m,uL1,lL1,uL2,lL2,uL3,lL3,sigma_SE,l_SE);
    
    %Simulated odometry measurement noise
    w=mvnrnd(zeros(3,1),sigma_p^2*eye(3),N-1)';
    
    %Generate fake odometry based on ground truth
    delta_p=gt(:,2:N)-gt(:,1:(N-1))+w+b;
    
    for noise_index=1:n_noises
        
        %Pre-allocate position trajectory for all particles
        p=zeros(3,N,M)+[lL1+margin; lL2+margin; z];
        
        %Initialise all weights to 1/M
        weights=ones(1,M)./M;
        
        %Pre-allocate magnetic field state vector for all particles
        m=zeros(N_m,M);
        
        %Pre-allocate the magnetic field variance for all particles
        P_m=zeros(N_m,N_m,M);
        for i=1:M
            P_m(:,:,i)=Lambda;
        end
        
        %Pre-allocate the dead-reckoning estimate of the position
        %based on the simulated odometry
        p_odometry=zeros(3,N)+[lL1+margin; lL2+margin; z];
        
        %iterate through all timesteps
        for t=2:N
            
            %perform dynamic update of dead reckoning position estimate
            p_odometry(:,t)=p_odometry(:,t-1)+delta_p(:,t-1);
            
            %perform dynamic update of all particles
            for i=1:M
                %Dynamic update
                p(:,t,i)=p(:,t-1,i)+delta_p(:,t-1)+mvnrnd(zeros(3,1),noises(noise_index)*sigma_p^2*eye(3))';
            end
            
            %Perform measurement update of all particles
            delay=10;
            if t>delay
                for i=1:M
                    
                    %Update the magnetic field maps based on the incoming
                    %measurements - Kalman filter measurement update
                    phi_t=Phi3D(p(:,t-delay,i),N_m,uL1,lL1,uL2,lL2,uL3,lL3,Indices);
                    S_t=phi_t*P_m(:,:,i)*phi_t'+sigma_y^2;
                    K_t=P_m(:,:,i)*phi_t'./S_t;
                    m(:,i)=m(:,i)+K_t*(y_sim(t-delay)-phi_t*m(:,i));
                    P_m(:,:,i)=P_m(:,:,i)-K_t*S_t*K_t';
                    P_m(:,:,i)=1/2*(P_m(:,:,i)+P_m(:,:,i)');
                    
                    
                    
                    %Evaluate weights of particles
                    phi_s=Phi3D(p(:,t,i),N_m,uL1,lL1,uL2,lL2,uL3,lL3,Indices);
                    mu=phi_s*m(:,i);
                    var=phi_s*P_m(:,:,i)*phi_s'+sigma_y^2;
                    weights(i) = normpdf(y_sim(t),mu,sqrt(var))*weights(i);
                    
                end
                
                %Normalise particle weights
                weights=weights./sum(weights);
                
                %Selective resampling of particles
                M_eff=1/sum(weights.^2);
                if (M_eff<(2*M/3))
                    [p,m,P_m]=resample(p,m,P_m,weights);
                    weights=0*weights+1./M;
                end
                
            end
            
        end
        
        
        RMSEsRBPF(experiment,noise_index)=calculate_RMSE(p,weights,gt)
        RMSEsODOMETRY(experiment,noise_index)=calculate_RMSE(p_odometry,1,gt)
        
    end
end
save('MCSimResults2.mat');

[max_val arg_max]=max(weights);
figure(2); clf;
plot3(p(2,:,arg_max),p(1,:,arg_max),p(3,:,arg_max),'k');
hold on;
plot3(p_odometry(2,:),p_odometry(1,:),p_odometry(3,:),'b');
plot3(gt(2,:),gt(1,:),gt(3,:),'r');
xlabel('$x/[m]$','Interpreter','Latex','Fontsize',14);
ylabel('$y/[m]$','Interpreter','Latex','Fontsize',14);
zlabel('$z/[m]$','Interpreter','Latex','Fontsize',14);
legend({'Mag-SLAM','Odometry','Ground truth'},'Interpreter','Latex','Fontsize',14);
axis equal;
grid on;